import pProps from 'p-props';
import assert from 'assert';
import { readFile, readFileSync } from 'fs';

/**
 * An error thrown when configuration could not be loaded.
 */
class ConfigError extends Error {
  /**
   * A collection of the individual errors, keyed by property name.
   */

  /**
   * Create a ConfigError.
   * @param errors - The errors that this ConfigError represent.
   */
  constructor(errors) {
    super(['Configuration could not be loaded for the following properties:', ...Object.entries(errors).map(([key, {
      message
    }]) => `${key}: ${message}`)].join('\n\t'));
    this.errors = errors;
  }
  /**
   * Convert this ConfigError to JSON.
   *
   * @remarks
   * This is here so the ConfigError can be stringified a little cleaner.  Throwing it'll still just show the base message.
   * @returns An object containing the message of this error and the messages of the errors it represents.
   */


  toJSON() {
    return {
      errors: Object.fromEntries(Object.entries(this.errors).map(([key, {
        message
      }]) => [key, message])),
      message: this.message
    };
  }

}

/**
 * Determine whether a PropertyConfig is an EnvironmentConfig.
 *
 * @param propertyConfig - The PropertyConfig that may be an EnvironmentConfig.
 * @param propertyName - The name of the property.
 * @returns True if the PropertyConfig is an EnvironmentConfig.
 */
function isEnvironmentConfig(propertyConfig, propertyName) {
  const isEnvironmentConfig = ('variableName' in propertyConfig);
  const isFileConfig = ('filePath' in propertyConfig);
  assert(isEnvironmentConfig || isFileConfig, `"configMap.${propertyName}" must be a string, EnvironmentConfig object, or FileConfig object.  Neither "filePath" nor "variableName" are defined.`);
  assert(!(isEnvironmentConfig && isFileConfig), `Cannot determine whether "configMap.${propertyName}" is an EnvironmentConfig object or a FileConfig object.  Both "filePath" and "variableName" are defined.`);
  return isEnvironmentConfig;
}

function tryToFormat(value, formatter) {
  try {
    return {
      error: false,
      value: formatter(value)
    };
  } catch (error) {
    return {
      error,
      value: undefined
    };
  }
}

function something(propertyResult, propertyConfig) {
  const {
    defaultValue,
    format,
    required = true
  } = propertyConfig;
  const hasDefaultValue = typeof defaultValue !== 'undefined';
  const hasError = propertyResult.error !== false;
  const hasFormat = typeof format !== 'undefined';
  const hasValue = typeof propertyResult.value !== 'undefined';

  if (hasError ? required && !hasDefaultValue : hasValue && !hasFormat) {
    return propertyResult;
  }

  return hasValue && hasFormat ? tryToFormat(propertyResult.value, format) : {
    error: false,
    value: defaultValue
  };
}

function loadEnvironmentProperty(propertyConfig) {
  const {
    variableName
  } = propertyConfig;
  const value = process.env[variableName];
  return something({
    error: typeof value !== 'string' && new Error(`${variableName} is not defined.`),
    value
  }, propertyConfig);
}

function loadFileProperty(propertyConfig) {
  const {
    encoding = 'utf8',
    filePath
  } = propertyConfig;
  return new Promise(resolve => {
    readFile(filePath, encoding, (error, value) => {
      resolve(something({
        error: error !== null && error !== void 0 ? error : false,
        value
      }, propertyConfig));
    });
  });
}

function loadProperty(propertyConfig, propertyName) {
  if (typeof propertyConfig === 'string') {
    return loadProperty({
      variableName: propertyConfig
    }, propertyName);
  }

  return isEnvironmentConfig(propertyConfig, propertyName) ? loadEnvironmentProperty(propertyConfig) : loadFileProperty(propertyConfig);
}

/**
 * Load configuration asynchronously.
 *
 * @example
 * ```
 * const config = await loadConfig({
 *   postgresPassword: {
 *     filePath: '/secrets/password',
 *   },
 *   postgresUrl: {
 *     format: (value) => new URL(value),
 *     variableName: 'POSTGRES_URL'
 *   },
 *   postgresUsername: {
 *     defaultValue: 'postgres',
 *     filePath: '/secrets/username',
 *   },
 * });
 *
 * assert(typeof config.postgresPassword === 'string');
 * assert(config.postgresUrl instanceof URL);
 * assert(typeof config.postgresUsername === 'string');
 * ```
 *
 * @param configMap - An object map of configuration properties and how they should be loaded.
 * @returns A Promise that resolves to the loaded configuration object.
 *
 * @throws {@link ConfigError}
 * Rejects if one or more configuration properties was not able to load properly.
 */
async function loadConfig(configMap) {
  const errorMap = new Map();
  const result = await pProps(configMap, async (propertyConfig, propertyName) => {
    const {
      error,
      value
    } = await loadProperty(propertyConfig, propertyName);

    if (error !== false) {
      errorMap.set(propertyName, error);
    }

    return value;
  });

  if (errorMap.size > 0) {
    const errors = Object.fromEntries(errorMap);
    throw new ConfigError(errors);
  }

  return result;
}

function tryToReadFile(propertyConfig) {
  const {
    encoding = 'utf8',
    filePath
  } = propertyConfig;

  try {
    return {
      error: false,
      value: readFileSync(filePath, encoding)
    };
  } catch (error) {
    return {
      error: error,
      value: undefined
    };
  }
}

function loadFilePropertySync(propertyConfig) {
  return something(tryToReadFile(propertyConfig), propertyConfig);
}

function loadPropertySync(propertyConfig, propertyName) {
  if (typeof propertyConfig === 'string') {
    return loadPropertySync({
      variableName: propertyConfig
    }, propertyName);
  }

  return isEnvironmentConfig(propertyConfig, propertyName) ? loadEnvironmentProperty(propertyConfig) : loadFilePropertySync(propertyConfig);
}

/**
 * Load configuration synchronously.
 *
 * @example
 * ```
 * const config = loadConfigSync({
 *   postgresPassword: {
 *     filePath: '/secrets/password',
 *   },
 *   postgresUrl: {
 *     format: (value) => new URL(value),
 *     variableName: 'POSTGRES_URL'
 *   },
 *   postgresUsername: {
 *     defaultValue: 'postgres',
 *     filePath: '/secrets/username',
 *   },
 * });
 *
 * assert(typeof config.postgresPassword === 'string');
 * assert(config.postgresUrl instanceof URL);
 * assert(typeof config.postgresUsername === 'string');
 * ```
 *
 * @param configMap - An object map of configuration properties and how they should be loaded.
 * @returns The loaded configuration object.
 *
 * @throws {@link ConfigError}
 * Thrown if one or more configuration properties was not able to load properly.
 */
function loadConfigSync(configMap) {
  const errorMap = new Map();
  const result = {};

  for (const [propertyName, propertyConfig] of Object.entries(configMap)) {
    const {
      error,
      value
    } = loadPropertySync(propertyConfig, propertyName);

    if (error !== false) {
      errorMap.set(propertyName, error);
    }

    result[propertyName] = value;
  }

  if (errorMap.size > 0) {
    const errors = Object.fromEntries(errorMap);
    throw new ConfigError(errors);
  }

  return result;
}

export { ConfigError, loadConfig, loadConfigSync };
